#include "CSVExporter.h"
#include <functional>
#include <cstdio>

//--------------------------------------------------------------------------------
void CSVExportOption::SetDefaultFromProject(GraphProject const &proj)
{
  exportDataX = true;
  exportHeader = true;

  dataXHeaderText = proj.GetPlotProperty().AxisXLabel();
  exportedData.clear();

  int dataCount = 0;
  GraphDataX dataX;

  for (int index = 0; index < proj.GetCount(); ++index)
  {
    GraphDataRefPtr graphData = proj.GetItem(index);

    if (0 == index)
    {
      dataX = graphData->GetDataX();
      dataCount = graphData->GetCount();
      
      CSVExportedDataInfo info;
      info.headerText = graphData->GetLabel();
      info.graphDataIndex = index;
      exportedData.push_back(info);
    }
    else
    {
      if (graphData->GetCount() == dataCount && graphData->GetDataX() == dataX)
      {
        CSVExportedDataInfo info;
        info.headerText = graphData->GetLabel();
        info.graphDataIndex = index;
        exportedData.push_back(info);
      }
    }
  }
}

#if !defined(NDEBUG)
void CSVExportOption::LogDebug() const
{
  wxLogDebug("exportDataX = %s", exportDataX ? "true": "false");
  wxLogDebug("exportHeader = %s", exportHeader ? "true": "false");
  wxLogDebug("dataXHeaderText = %s", dataXHeaderText);
  for (size_t i = 0; i < exportedData.size(); ++i)
  {
    CSVExportedDataInfo info = exportedData[i];
    wxLogDebug("exportedData[%d]: graphDataIndex = %d, headerText = %s",
      i, info.graphDataIndex, info.headerText);
  }
}
#endif

//--------------------------------------------------------------------------------
CSVExporter::CSVExporter(wxString const &fileName, GraphProject &project)
  : m_project(project),
    m_fileName(fileName),
    m_separator(_T(','))
{
  m_option.SetDefaultFromProject(project);
}

void CSVExporter::SetOption(CSVExportOption const &option)
{
  m_option = option;
}

void CSVExporter::GetOption(CSVExportOption &option) const
{
  option = m_option;
}

void CSVExporter::SetSeparator(wxUniChar separator)
{
  m_separator = separator;
}

wxUniChar CSVExporter::GetSeparator() const
{
  return m_separator;
}

ResultCode CSVExporter::Export(ProcContext *pPC)
{
  using namespace std::placeholders;

  ResultCode result = RunProcWithContext<ResultCode>
    (std::bind(&CSVExporter::ExportInternal, this,  _1), pPC);

  return result;
}

ResultCode CSVExporter::ExportInternal(Progress *pProgress)
{
  if (!IsOptionValid())
  {
    return ResultCode::INVALID_PARAM;
  }
  
  char sep;
  bool b = m_separator.GetAsChar(&sep);
  wxASSERT(b);

  // I[v
  std::string fileNameLC = m_fileName.ToStdString();  // current locale
  FILE *file = ::fopen(fileNameLC.c_str(), "w");
  if (NULL == file)
  {
    m_err.Printf(_("Can't create \"%s\"!"), m_fileName);
    return ResultCode::FILE_ERROR;
  }
  
  // 񐔂̊mF
  int colCount = m_option.exportedData.size();
  
  // wb_[
  if (m_option.exportHeader)
  {
    if (m_option.exportDataX)
    {
      wxString headerText = m_option.dataXHeaderText;
      std::string headerTextLC = headerText.ToStdString();  // current locale
      ::fprintf(file, "%s%c", headerTextLC.c_str(), sep);
    }
    for (int col = 0; col < colCount; ++col)
    {
      wxString headerText = m_option.exportedData[col].headerText;
      std::string headerTextLC = headerText.ToStdString();  // current locale
      if (col != 0) ::fprintf(file, "%c", sep);
      ::fprintf(file, "%s", headerTextLC.c_str());
    }
    ::fprintf(file, "\n");
  }

  // ݃f[^̗p
  GraphDataX dataX;
  std::vector<GraphDataYRefPtr> dataYArray;
  dataYArray.resize(colCount);
  for (int col = 0; col < colCount; ++col)
  {
    int index = m_option.exportedData[col].graphDataIndex;
    wxASSERT(index >= 0 && index < m_project.GetCount());
    GraphDataRefPtr graphData = m_project.GetItem(index);

    if (0 == col)
    {
      dataX = graphData->GetDataX();
    }
    else
    {
      wxASSERT(graphData->GetDataX() == dataX);
    }
    dataYArray[col] = graphData->GetDataY();
  }

  wxASSERT(!dataYArray.empty());
  int rowCount = dataYArray[0]->raw.size();
  
  // f[^
  int updateInterval = rowCount / 500;
  if (0 == updateInterval) updateInterval = 1;
  if (pProgress != NULL)
  {
    wxString message;
    message.Printf(_("Export \"%s\" ..."), m_fileName);
    pProgress->UpdateMessage(message);
  }

  if (m_option.exportDataX)
  {
    double x = dataX.begin;
    for (int row = 0; row < rowCount; ++row, x += dataX.interval)
    {
      ::fprintf(file, "%.8e", x);
      for (int col = 0; col < colCount; ++col)
      {
        double val = dataYArray[col]->raw[row];
        ::fprintf(file, "%c%.8e", sep, val);
      }
      ::fprintf(file, "\n");

      // vOXo[XV.
      if (0 == row % updateInterval && pProgress != NULL)
      {
        pProgress->UpdateProgress(row, rowCount);
      }

      // fmF
      if (0 == row % 1000 && pProgress != NULL && pProgress->IsAbortRequested())
      {
        ::fclose(file);
        ::remove(fileNameLC.c_str());
        return ResultCode::ABORT;
      }
    }
  }
  else
  {
    for (int row = 0; row < rowCount; ++row)
    {
      double val = dataYArray[0]->raw[row];
      ::fprintf(file, "%.8e", val);
      for (int col = 1; col < colCount; ++col)
      {
        val = dataYArray[col]->raw[row];
        ::fprintf(file, "%c%.8e", sep, val);
      }
      ::fprintf(file, "\n");

      // vOXo[XV.
      if (0 == row % updateInterval && pProgress != NULL)
      {
        pProgress->UpdateProgress(row, rowCount);
      }

      // fmF
      if (0 == row % 1000 && pProgress != NULL && pProgress->IsAbortRequested())
      {
        ::fclose(file);
        ::remove(fileNameLC.c_str());
        return ResultCode::ABORT;
      }
    }
  }

  // N[Y
  ::fclose(file);

  return ResultCode::SUCCESS;
}

wxString CSVExporter::GetLastErrorMessage() const
{
  return m_err;
}

bool CSVExporter::IsOptionValid() const
{
  if (m_option.exportedData.empty())
  {
    m_err = _("Exported data is empty!");
    return false;
  }

  int dataCount = 0;
  GraphDataX dataX;

  int colCount = m_option.exportedData.size();
  for (int col = 0; col < colCount; ++col)
  {
    int index = m_option.exportedData[col].graphDataIndex;
    GraphDataRefPtr graphData = m_project.GetItem(index);

    if (0 == col)
    {
      dataX = graphData->GetDataX();
      dataCount = graphData->GetCount();
    }
    else
    {
      if (graphData->GetCount() != dataCount || graphData->GetDataX() != dataX)
      {
        m_err = _("Exported data doesn't match x data or size of data!");
        return false;
      }
    }
  }

  return true;
}

//--------------------------------------------------------------------------------
