#include "Cairo.h"
#include <cairo-pdf.h>
#include <cairo-ps.h>
#include <cairo-svg.h>

//--------------------------------------------------------------------------------
void ClosedPath::Expand(double h)
{
  int N = m_points.size();
  if (N <= 2) return;

  std::vector<wxPoint2DDouble> q, v, d, r;
  q.resize(N);
  v.resize(N);
  d.resize(N);
  r.resize(N);

  // xNg v ߂.
  for (int i = 0; i < N - 1; ++i)
  {
    v[i] = m_points[i + 1] - m_points[i];
  }
  v[N - 1] = m_points[0] - m_points[N - 1];

  // _ q ߂.
  for (int i = 0; i < N; ++i)
  {
    double vx = v[i].m_x;
    double vy = v[i].m_y;
    double norm = std::sqrt(vx * vx + vy * vy);
    wxPoint2DDouble n(vy / norm, -vx / norm);
    q[i] = m_points[i] + h * n;
  }

  // xNg d ߂.
  for (int i = 0; i < N - 1; ++i)
  {
    d[i] = q[i] - q[i + 1];
  }
  d[N - 1] = q[N - 1] - q[0];

  // g又
  int j = 0;
  for (int i = 0; i < N - 1; ++i)
  {
    double t_num = v[i + 1].m_y * v[i].m_x - v[i + 1].m_x * v[i].m_y;
    if (t_num != 0.0)
    {
      double t = (v[i + 1].m_x * d[i].m_y - v[i + 1].m_y * d[i].m_x) / t_num;
      r[j] = q[i] + t * v[i];
      ++j;
    }
  }

  double t_num = v[0].m_y * v[N - 1].m_x - v[0].m_x * v[N - 1].m_y;
  if (t_num != 0.0)
  {
    double t = (v[0].m_x * d[N - 1].m_y - v[0].m_y * d[N - 1].m_x) / t_num;
    r[j] = q[N - 1] + t * v[N - 1];
    ++j;
  }
  
  r.resize(j);

  m_points.swap(r);
}

//--------------------------------------------------------------------------------
Cairo::Cairo(HDC dc)
  : m_dc(dc),
    m_surface(NULL),
    m_cr(NULL),
    m_type(TYPE_DC)
{
  m_surface = ::cairo_win32_surface_create(dc);
  m_cr = ::cairo_create(m_surface);
  if (::cairo_status(m_cr) == CAIRO_STATUS_NO_MEMORY)
  {
    ::cairo_surface_destroy(m_surface); m_surface = NULL;
    ::cairo_destroy(m_cr); m_cr = NULL;
    m_type = TYPE_INVALID;
  }
}

Cairo::Cairo(wxString const &fileName, double width, double height)
  : m_dc(NULL),
    m_surface(NULL),
    m_cr(NULL),
    m_type(TYPE_INVALID),
    m_fileName(fileName)
{
  if (_T(".PDF") == m_fileName.Right(4).MakeUpper())
  {
    m_surface = ::cairo_pdf_surface_create(m_fileName.ToUTF8(), width, height);
    m_type = TYPE_PDF;
  }
  else if (_T(".PS") == m_fileName.Right(3).MakeUpper())
  {
    m_surface = ::cairo_ps_surface_create(m_fileName.ToUTF8(), width, height);
    m_type = TYPE_PS;
  }
  else if (_T(".SVG") == m_fileName.Right(4).MakeUpper())
  {
    m_surface = ::cairo_svg_surface_create(m_fileName.ToUTF8(), width, height);
    m_type = TYPE_SVG;
  }
  else if (_T(".PNG") == m_fileName.Right(4).MakeUpper())
  {
    m_surface = ::cairo_image_surface_create(CAIRO_FORMAT_ARGB32, static_cast<int>(width), static_cast<int>(height));
    m_type = TYPE_PNG;
  }
  
  if (m_surface == NULL)
  {
    m_type = TYPE_INVALID;
  }
  else
  {
    m_cr = ::cairo_create(m_surface);
    if (::cairo_status(m_cr) == CAIRO_STATUS_NO_MEMORY)
    {
      ::cairo_surface_destroy(m_surface); m_surface = NULL;
      ::cairo_destroy(m_cr); m_cr = NULL;
      m_type = TYPE_INVALID;
    }
  }
}

Cairo::~Cairo()
{
  if (m_type == TYPE_PNG)
  {
    ::cairo_surface_write_to_png(m_surface, m_fileName.ToUTF8());
  }
  else if (m_type == TYPE_PS || m_type == TYPE_PDF)
  {
    ::cairo_surface_show_page(m_surface);
  }
  
  if (m_surface != NULL) ::cairo_surface_destroy(m_surface);
  if (m_cr != NULL) ::cairo_destroy(m_cr);
}

void Cairo::SetSourceRGBA(wxColour const &color)
{
  double r = static_cast<double>(color.Red())   / 255.0;
  double g = static_cast<double>(color.Green()) / 255.0;
  double b = static_cast<double>(color.Blue())  / 255.0;
  double a = static_cast<double>(color.Alpha()) / 255.0;
  ::cairo_set_source_rgba(m_cr, r, g, b, a);
}


void Cairo::SetFont(wxFont const &font)
{
  cairo_font_slant_t slant;
  switch (font.GetStyle())
  {
  case wxFONTSTYLE_ITALIC:
  case wxFONTSTYLE_SLANT:
    slant = CAIRO_FONT_SLANT_ITALIC;
    break;
  case wxFONTSTYLE_NORMAL:
  default:
    slant = CAIRO_FONT_SLANT_NORMAL;
  }

  cairo_font_weight_t weight;
  switch (font.GetWeight())
  {
  case wxFONTWEIGHT_BOLD:
    weight = CAIRO_FONT_WEIGHT_BOLD;
    break;
  case wxFONTWEIGHT_NORMAL:
  case wxFONTWEIGHT_LIGHT:
  default:
    weight = CAIRO_FONT_WEIGHT_NORMAL;
  }

  wxString faceName = font.GetFaceName();
  ::cairo_select_font_face(m_cr, faceName.utf8_str(), slant, weight);

  double size = font.GetPixelSize().GetY();
  ::cairo_set_font_size(m_cr, size);
}


void Cairo::ShowAlignText(double x, double y, double width, double height,
  double alignX, double alignY, wxString const &text)
{
  wxASSERT(m_type != TYPE_INVALID);
  ::cairo_save(m_cr);

  cairo_text_extents_t ext;
  ::cairo_text_extents(m_cr, text.ToUTF8(), &ext);
  ::cairo_move_to(m_cr,
    -ext.x_bearing + x + (width - ext.width) * alignX,
    -ext.y_bearing + y + (height - ext.height) * alignY);
  ::cairo_show_text(m_cr, text.ToUTF8());
  
  ::cairo_restore(m_cr);
}

void Cairo::ShowText(wxString const &text)
{
  wxASSERT(m_type != TYPE_INVALID);

  ::cairo_show_text(m_cr, text.ToUTF8());
}

void Cairo::ShowPalette(double x, double y, double width, double height, COLORREF *pPalette, int paletteSize)
{
  wxASSERT(m_type != TYPE_INVALID);
  
  cairo_pattern_t *pat = ::cairo_pattern_create_linear(x, 0.0, x + width, 0.0);
  for (int i = 0; i < paletteSize; ++i)
  {
    unsigned int r = GetRValue(pPalette[i]);
    unsigned int g = GetGValue(pPalette[i]);
    unsigned int b = GetBValue(pPalette[i]);
    ::cairo_pattern_add_color_stop_rgb(pat, static_cast<double>(i) / (paletteSize - 1),
      r / 255.0, g / 255.0, b / 255.0);
  }
  ::cairo_rectangle(m_cr, x, y, width, height);
  ::cairo_set_source(m_cr, pat);
  ::cairo_fill(m_cr);
  ::cairo_pattern_destroy(pat);
}

void Cairo::AppendClosedPath(ClosedPath const &path)
{
  std::vector<wxPoint2DDouble> const &points = path.Points();
  std::vector<wxPoint2DDouble>::const_iterator i = points.begin();
  if (i != points.end())
  {
    ::cairo_move_to(m_cr, i->m_x, i->m_y);
    ++i;
  }
  for ( ; i != points.end(); ++i)
  {
    ::cairo_line_to(m_cr, i->m_x, i->m_y);
  }
  ::cairo_close_path(m_cr);
}

//--------------------------------------------------------------------------------
