#include "GraphRenderer.h"

#ifdef NDEBUG
  #define DEFINE_STOPWATCH()
  #define START_STOPWATCH()
  #define STOP_STOPWATCH(str)
#else
  #define DEFINE_STOPWATCH() \
    wxLongLong stopwatchStart = 0;
  #define START_STOPWATCH() \
    stopwatchStart = wxGetLocalTimeMillis();
  #define STOP_STOPWATCH(str) \
    wxLogDebug("Stopwatch: %s = %f ms", str, (wxGetLocalTimeMillis() - stopwatchStart).ToDouble());
#endif

//--------------------------------------------------------------------------------
GraphRenderingLayout::GraphRenderingLayout()
  : clientWidth(1.0), clientHeight(1.0),
    rectGraphFrame(0.0, 0.0, 1.0, 1.0),
    rangeMinX(0.0), rangeMaxX(1.0),
    rangeMinY(0.0), rangeMaxY(1.0),
    gridIntervalX(0.0), gridIntervalY(0.0)
{
}

wxPoint2DDouble GraphRenderingLayout::TransformClientToGraphPoint(wxPoint const &pt) const
{
  double graphFrameLeft   = rectGraphFrame.GetLeft();
  double graphFrameBottom = rectGraphFrame.GetBottom();
  double graphFrameWidth  = rectGraphFrame.GetRight() - rectGraphFrame.GetLeft();
  double graphFrameHeight = rectGraphFrame.GetBottom() - rectGraphFrame.GetTop();

  double graphRangeWidth  = rangeMaxX - rangeMinX;
  double graphRangeHeight = rangeMaxY - rangeMinY;

  double x = rangeMinX + (pt.x - graphFrameLeft) * graphRangeWidth / graphFrameWidth;
  double y = rangeMinY + (graphFrameBottom - pt.y) * graphRangeHeight / graphFrameHeight;
  
  return wxPoint2DDouble(x, y);
}

wxPoint2DDouble GraphRenderingLayout::TransformClientToGraphDistance(wxPoint const &pt) const
{
  double graphFrameWidth  = rectGraphFrame.GetRight() - rectGraphFrame.GetLeft();
  double graphFrameHeight = rectGraphFrame.GetBottom() - rectGraphFrame.GetTop();

  double graphRangeWidth  = rangeMaxX - rangeMinX;
  double graphRangeHeight = rangeMaxY - rangeMinY;

  double x = pt.x * graphRangeWidth / graphFrameWidth;
  double y = -pt.y * graphRangeHeight / graphFrameHeight;

  return wxPoint2DDouble(x, y);
}

//--------------------------------------------------------------------------------
GraphRenderer::GraphRenderer(Cairo &cairo, GraphProject &project)
  : m_cairo(cairo),
    m_proj(project),
    m_width(0.0), m_height(0.0),
    m_resolution(-1), m_accuracy(1)
{
}

void GraphRenderer::SetArea(double width, double height)
{
  m_width = width;
  m_height = height;
}

void GraphRenderer::SetResolution(int resolution, int accuracy)
{
  m_resolution = resolution;
  m_accuracy = accuracy;
}

void GraphRenderer::Render()
{
  wxASSERT(m_cairo() != NULL);
  wxASSERT(m_width != 0.0 && m_height != 0.0);

  DEFINE_STOPWATCH();
  START_STOPWATCH();

  // Cairo ̐ݒ
  //::cairo_set_antialias(m_cairo(), CAIRO_ANTIALIAS_FAST);
  ::cairo_set_line_join(m_cairo(), CAIRO_LINE_JOIN_ROUND);

  // vbĝ߂̐ݒ擾
  GraphProjectPlotProperty projProp = m_proj.GetPlotProperty();
  //projProp.SetGraphArea(m_width, m_height);

  PlotRange r;
  m_proj.GetPlotRange(r.minX, r.maxX, r.minY, r.maxY);

  // wihԂ
  ::cairo_rectangle(m_cairo(), 0.0, 0.0, m_width, m_height);
  m_cairo.SetSourceRGBA(projProp.BgColor());
  ::cairo_fill(m_cairo());

  STOP_STOPWATCH("background");
  START_STOPWATCH();

  // Oẗ
  ::cairo_save(m_cairo());
    double gw = m_width - projProp.MarginLeft() - projProp.MarginRight();
    double gh = m_height - projProp.MarginTop() - projProp.MarginBottom();
    ::cairo_translate(m_cairo(), projProp.MarginLeft(), projProp.MarginTop());

    if (m_resolution <= 0)
    {
      // \
      m_resolution = static_cast<int>(gw);
    }
    m_resolution /= m_accuracy;

    // Otf[^`f[^ւ̍Wϊp[^߂
    STMatrix m;
    m.sx = gw / (r.maxX - r.minX);  m.sy = -gh / (r.maxY - r.minY);
    m.tx = -r.minX * m.sx;          m.ty = -r.minY * m.sy + gh;
    
    // Obh̃vbg
    RenderGrid(projProp, r, m, gw, gh);
    
    STOP_STOPWATCH("grid");
    START_STOPWATCH();

    ::cairo_save(m_cairo());
      // Nbv[W̐ݒ
      ::cairo_rectangle(m_cairo(), 0.0, 0.0, gw, gh);
      ::cairo_clip(m_cairo());

      // Otf[^̃vbg
      RenderData(r, m);
    ::cairo_restore(m_cairo());

    STOP_STOPWATCH("data");
    START_STOPWATCH();

    // ڐ, ڐ胉x̃vbg
    RenderTick(projProp, r, m, gw, gh);
    
    // Oẗ̘g̕`
    ::cairo_rectangle(m_cairo(), 0.0, 0.0, gw, gh);
    ::cairo_set_line_width(m_cairo(), projProp.GraphAreaFrameWidth());
    m_cairo.SetSourceRGBA(projProp.GraphAreaFrameColor());
    ::cairo_stroke(m_cairo());

    // X, Y x̕`
    RenderAxisLabel(projProp, gw, gh);
  ::cairo_restore(m_cairo());

  STOP_STOPWATCH("labels");

  // G[`FbN
  cairo_status_t status = ::cairo_status(m_cairo());
  if (status != CAIRO_STATUS_SUCCESS)
  {
    wxLogDebug("Error occurs with using Cairo: %s\n", ::cairo_status_to_string(status));
  }

  // CAEg
  m_layout.clientWidth = m_width;
  m_layout.clientHeight = m_height;
  m_layout.rectGraphFrame = wxRect2DDouble(projProp.MarginLeft(), projProp.MarginTop(), gw, gh);
  m_layout.rangeMinX = r.minX;
  m_layout.rangeMaxX = r.maxX;
  m_layout.rangeMinY = r.minY;
  m_layout.rangeMaxY = r.maxY;
}

void GraphRenderer::GetLayout(GraphRenderingLayout &layout)
{
  layout = m_layout;
}

void GraphRenderer::RenderGrid(GraphProjectPlotProperty const &projProp,
    PlotRange const &r, STMatrix const &m, double gw, double gh)
{
  double intervalX = projProp.CalcGridIntervalX(m_proj, gw);
  double intervalY = projProp.CalcGridIntervalY(m_proj, gh);

  if (projProp.GridXEnabled() && intervalX > 0.0)
  {
    ::cairo_new_path(m_cairo());
    double x = intervalX * std::ceil(r.minX / intervalX);
    while (x <= r.maxX)
    {
      ::cairo_new_sub_path(m_cairo());
      ::cairo_move_to(m_cairo(), m.ConvX(x), 0.0);
      ::cairo_line_to(m_cairo(), m.ConvX(x), gh);
      x += intervalX;
    }
    ::cairo_set_line_width(m_cairo(), projProp.GridLineWidth());
    m_cairo.SetSourceRGBA(projProp.GridLineColor());
    ::cairo_stroke(m_cairo());
  }

  if (projProp.GridYEnabled() && intervalY > 0.0)
  {
    ::cairo_new_path(m_cairo());
    double y = intervalY * std::ceil(r.minY / intervalY);
    while (y <= r.maxY)
    {
      ::cairo_new_sub_path(m_cairo());
      ::cairo_move_to(m_cairo(), 0.0, m.ConvY(y));
      ::cairo_line_to(m_cairo(), gw, m.ConvY(y));
      y += intervalY;
    }
    ::cairo_set_line_width(m_cairo(), projProp.GridLineWidth());
    m_cairo.SetSourceRGBA(projProp.GridLineColor());
    ::cairo_stroke(m_cairo());
  }
}

void GraphRenderer::RenderData(PlotRange const &r, STMatrix const &m)
{
  for (int i = 0; i < m_proj.GetCount(); ++i)
  {
    GraphDataRefPtr graph = m_proj.GetItem(i);
    if (graph->IsEnabled())
    {
      PlotData plot;
      int points = graph->GetPlotDataInRangeX(plot, r.minX, r.maxX, m_resolution);
      if (points >= 2)
      {
        GraphDataPlotProperty dataProp = graph->GetPlotProperty();

        if (plot.useRaw)
        {
          // RAW f[^vbg
          double x = plot.dataX.begin;
          PlotData::RawIter iter = plot.beginRaw;
          ::cairo_new_path(m_cairo());
          ::cairo_move_to(m_cairo(), m.ConvX(x), m.ConvY(*iter));
          ++iter;
          x += plot.dataX.interval;
          for ( ; iter != plot.endRaw; ++iter, x += plot.dataX.interval)
          {
            ::cairo_line_to(m_cairo(), m.ConvX(x), m.ConvY(*iter));
          }
          ::cairo_set_line_width(m_cairo(), dataProp.LineWidth());
          m_cairo.SetSourceRGBA(dataProp.LineColor());
          ::cairo_stroke(m_cairo());
        }
        else
        {
          // ~bv}bvvbg
          double x = plot.dataX.begin;
          PlotData::MipmapIter iter = plot.beginMip;
#if 0
          // y\ɂ肠
          ClosedPath path;
          for ( ; iter != plot.endMip; ++iter, x += plot.dataX.interval)
          {
            path.AddPoint(m.ConvX(x), m.ConvY(iter->upper));
          }
          iter = plot.endMip - 1;
          x = plot.dataX.begin + plot.dataX.interval * (points - 1);
          for ( ; ; --iter, x -= plot.dataX.interval)
          {
            path.AddPoint(m.ConvX(x), m.ConvY(iter->lower));
            if (iter == plot.beginMip) break;
          }

          path.Expand(dataProp.lineW * 0.5);

          ::cairo_new_path(m_cairo());
          path.AppendToCairo(m_cairo());
          ::CairoSetSourceRGBA(m_cairo(), dataProp.lineColor);
          ::cairo_fill(m_cairo());

#elif 0
          // dȂ̂ŎgpȂ
          double prev_x = x;
          PlotData::MipmapIter prev_iter = iter;
          x += plot.dataX.interval;
          ++iter;

          double halfLineW = dataProp.lineW * 0.5;

          ::cairo_new_path(m_cairo());
          for ( ; iter != plot.endMip; ++iter, x += plot.dataX.interval)
          {
            ::cairo_new_sub_path(m_cairo());
            ::cairo_move_to(m_cairo(), m.ConvX(prev_x), m.ConvY(prev_iter->upper));
            ::cairo_line_to(m_cairo(), m.ConvX(x),      m.ConvY(iter->upper));
            ::cairo_line_to(m_cairo(), m.ConvX(x),      m.ConvY(iter->lower));
            ::cairo_line_to(m_cairo(), m.ConvX(prev_x), m.ConvY(prev_iter->lower));
            ::cairo_close_path(m_cairo());

            prev_x = x;
            prev_iter = iter;
          }

          ::cairo_set_line_width(m_cairo(), dataProp.lineW);
          ::CairoSetSourceRGBA(m_cairo(), dataProp.lineColor);
          ::cairo_stroke_preserve(m_cairo());
          ::cairo_fill(m_cairo());
#else
          // d\̖͂Ȃ
          ::cairo_new_path(m_cairo());
          ::cairo_move_to(m_cairo(), m.ConvX(x), m.ConvY(iter->upper));
          ++iter;
          x += plot.dataX.interval;
          for ( ; iter != plot.endMip; ++iter, x += plot.dataX.interval)
          {
            ::cairo_line_to(m_cairo(), m.ConvX(x), m.ConvY(iter->upper));
          }
          iter = plot.endMip - 1;
          x = plot.dataX.begin + plot.dataX.interval * (points - 1);
          for ( ; ; --iter, x -= plot.dataX.interval)
          {
            ::cairo_line_to(m_cairo(), m.ConvX(x), m.ConvY(iter->lower));
            if (iter == plot.beginMip) break;
          }
          ::cairo_close_path(m_cairo());
          
          ::cairo_set_line_width(m_cairo(), dataProp.LineWidth());
          m_cairo.SetSourceRGBA(dataProp.LineColor());
          ::cairo_stroke_preserve(m_cairo());
          ::cairo_fill(m_cairo());
#endif
        }
      }
    }
  }
}

void GraphRenderer::RenderTick(GraphProjectPlotProperty const &projProp,
    PlotRange const &r, STMatrix const &m, double gw, double gh)
{
  double intervalX = projProp.CalcGridIntervalX(m_proj, gw);
  double intervalY = projProp.CalcGridIntervalY(m_proj, gh);
  double len = projProp.TickLineLength();
  wxString formatX = projProp.TickLabelFormatX();
  wxString formatY = projProp.TickLabelFormatY();

  // ڐ̃vbg
  ::cairo_new_path(m_cairo());

  {
    double x = intervalX * std::ceil(r.minX / intervalX);
    while (x <= r.maxX)
    {
      ::cairo_new_sub_path(m_cairo());
      ::cairo_move_to(m_cairo(), m.ConvX(x), 0.0);
      ::cairo_line_to(m_cairo(), m.ConvX(x), len);
      ::cairo_new_sub_path(m_cairo());
      ::cairo_move_to(m_cairo(), m.ConvX(x), gh - len);
      ::cairo_line_to(m_cairo(), m.ConvX(x), gh);
      x += intervalX;
    }
  }
  
  {
    double y = intervalY * std::ceil(r.minY / intervalY);
    while (y <= r.maxY)
    {
      ::cairo_new_sub_path(m_cairo());
      ::cairo_move_to(m_cairo(), 0.0, m.ConvY(y));
      ::cairo_line_to(m_cairo(), len, m.ConvY(y));
      ::cairo_new_sub_path(m_cairo());
      ::cairo_move_to(m_cairo(), gw - len, m.ConvY(y));
      ::cairo_line_to(m_cairo(), gw, m.ConvY(y));
      y += intervalY;
    }
  }

  ::cairo_set_line_width(m_cairo(), projProp.TickLineWidth());
  m_cairo.SetSourceRGBA(projProp.TickLineColor());
  ::cairo_stroke(m_cairo());

  // ڐ胉x̃vbg
  m_cairo.SetFont(projProp.TickLabelFont());
  m_cairo.SetSourceRGBA(projProp.TickLabelFontColor());

  {
    double x = intervalX * std::ceil(r.minX / intervalX);
    
    double yy = -projProp.TickLabelPosY();
    if (projProp.TickLabelPosY() <= 0) yy += gh;
    
    while (x <= r.maxX)
    {
      wxString label;
      label.Printf(formatX, x);

      cairo_text_extents_t ext;
      ::cairo_text_extents(m_cairo(), label.ToUTF8(), &ext);
      ::cairo_move_to(m_cairo(), -ext.x_bearing + m.ConvX(x) - ext.width * 0.5, yy);
      ::cairo_show_text(m_cairo(), label.ToUTF8());

      x += intervalX;
    }
  }

  {
    double y = intervalY * std::ceil(r.minY / intervalY);
    
    double xx = projProp.TickLabelPosX();
    if (projProp.TickLabelPosX() > 0) xx += gw;

    while (y <= r.maxY)
    {
      wxString label;
      label.Printf(formatY, y);

      cairo_text_extents_t ext;
      ::cairo_text_extents(m_cairo(), label.ToUTF8(), &ext);
      ::cairo_move_to(m_cairo(), xx, -ext.y_bearing + m.ConvY(y) - ext.height * 0.5);
      ::cairo_show_text(m_cairo(), label.ToUTF8());

      y += intervalY;
    }
  }

  // CAEg
  m_layout.gridIntervalX = intervalX;
  m_layout.gridIntervalY = intervalY;
}

void GraphRenderer::RenderAxisLabel(GraphProjectPlotProperty const &projProp, double gw, double gh)
{
  cairo_text_extents_t ext;

  m_cairo.SetFont(projProp.AxisLabelFont());
  m_cairo.SetSourceRGBA(projProp.AxisLabelFontColor());

  // X x̕`
  double yy = -projProp.AxisXLabelPosY();
  if (projProp.AxisXLabelPosY() <= 0) yy += gh;

  ::cairo_text_extents(m_cairo(), projProp.AxisXLabel().ToUTF8(), &ext);
  ::cairo_move_to(m_cairo(), -ext.x_bearing + gw * 0.5 - ext.width * 0.5, yy);
  ::cairo_show_text(m_cairo(), projProp.AxisXLabel().ToUTF8());

  // Y x̕`
  double xx = projProp.AxisYLabelPosX();
  if (projProp.AxisYLabelPosX() > 0) xx += gw;

  ::cairo_save(m_cairo());
    ::cairo_translate(m_cairo(), xx, gh * 0.5);
    ::cairo_rotate(m_cairo(), -M_PI / 2.0);

    ::cairo_text_extents(m_cairo(), projProp.AxisYLabel().ToUTF8(), &ext);
    ::cairo_move_to(m_cairo(), -ext.x_bearing - ext.width * 0.5, 0.0);
    ::cairo_show_text(m_cairo(), projProp.AxisYLabel().ToUTF8());
  ::cairo_restore(m_cairo());
}

//--------------------------------------------------------------------------------
